//------------------------------------------------------------------------------
//
// render.cpp
//
//------------------------------------------------------------------------------
#include "common.h"

#include "crc32.h"

#include <vector>

//------------------------------------------------------------------------------
// Struct 
//------------------------------------------------------------------------------
struct D4_DIFF_TEX {
  FLOAT x, y, z, w;
  DWORD color;
  FLOAT u, v;
  enum { FVF = D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1 };
};

struct D3_DIFF_TEX {
  FLOAT x, y, z;
  DWORD color;
  FLOAT u, v;
  enum { FVF = D3DFVF_XYZ | D3DFVF_DIFFUSE | D3DFVF_TEX1 };
};


//Mesh
struct Mesh {
  LPD3DXMESH              lpMesh;
  LPD3DXBUFFER            lpAdjacency;
  D3DMATERIAL9            *lpMeshMaterial;
  LPDIRECT3DTEXTURE9      *lpMeshTexture;
  DWORD                   dwMaterialCount;
};

//Loadable Texture
struct TextureInfo {
  unsigned long      crc32;
  LPDIRECT3DTEXTURE9 lpTex;
  int                nRef;
};


//------------------------------------------------------------------------------
// DirectX OBJECT
//------------------------------------------------------------------------------

struct RenderData {
  LPDIRECT3D9              lpD3D;
  LPDIRECT3DDEVICE9        lpD3DDevice;
  D3DPRESENT_PARAMETERS    d3dpp;
  D3DADAPTER_IDENTIFIER9   d3daid;
  D3DDISPLAYMODE           d3ddm;

  //------------------------------------------------------------------------------
  // DirectX RESOURCE
  //------------------------------------------------------------------------------
  //Old Back Surface
  LPDIRECT3DSURFACE9       lpBackSurface;
  LPDIRECT3DSURFACE9       lpBackZBuffer;

  //Flip Surface
  BOOL                     bUseBlur;
  LPDIRECT3DTEXTURE9       lpFlipTexture[FlipTextureMax];
  LPDIRECT3DSURFACE9       lpFlipSurface[FlipTextureMax];
  LPDIRECT3DSURFACE9       lpFlipZBuffer[FlipTextureMax];
  DWORD                    FlipNumber;

  //Effect Texture
  LPDIRECT3DTEXTURE9       lpEffectTexture;
  LPDIRECT3DSURFACE9       lpEffectSurface;
  D3DCOLOR                 dwColorBlur;

  //Static Font
  LPD3DXFONT               lpFontObj[FontMax];
  LPD3DXSPRITE             lpFontSprite;
  
  //Mesh
  Mesh                     meshobj[MeshMax];
  TextureInfo              texinfo[TextureMax];
  
  //Render State
  DWORD                    dwRecordBits;
  
  //COnfigration
  BOOL                     bUseMipMap;
};

//staticobj
static RenderData render;

//macro
#define Device             render.lpD3DDevice

//------------------------------------------------------------------------------
// FUNCTION (Font)
//------------------------------------------------------------------------------
HRESULT ReleaseFontData(DWORD dwNum);

HRESULT CreateFontData(DWORD dwNum, LPCSTR szName, UINT pitch) {
  HRESULT hRet = E_FAIL;
  SystemParametersInfo(SPI_SETFONTSMOOTHING, TRUE, NULL, 0 );
  if(FontMax < dwNum) return hRet;
  if(strlen(szName) >= LF_FACESIZE) {
    ShowErrorMessage("CreateFontData : Invalid Font Face\n", "Error");
    return E_FAIL;
  }
  
  if(Device) {
    if(!render.lpFontObj[dwNum]) {
      if(!render.lpFontSprite) {
        if( FAILED( D3DXCreateSprite(Device, &render.lpFontSprite) ) ) {
          ShowErrorMessage("CreateFontData : D3DXCreateSprite(render.lpFontSprite)\n", "Error");
          ReleaseFontData(dwNum);
          return E_FAIL;
        }
      }
      
      D3DXFONT_DESC d3dxfdesc;
      d3dxfdesc.Height          =    pitch;
      d3dxfdesc.Width           =    0;
      d3dxfdesc.Weight          =    FW_NORMAL;
      d3dxfdesc.MipLevels       =    1;
      d3dxfdesc.Italic          =    FALSE;
      d3dxfdesc.CharSet         =    DEFAULT_CHARSET;
      d3dxfdesc.OutputPrecision =    OUT_OUTLINE_PRECIS;
      d3dxfdesc.Quality         =    PROOF_QUALITY;
      d3dxfdesc.PitchAndFamily  =    DEFAULT_PITCH | FF_DONTCARE; 
      wsprintf(d3dxfdesc.FaceName, "%s", szName);
      if( FAILED( D3DXCreateFontIndirect(Device, &d3dxfdesc, &render.lpFontObj[dwNum]) ) ) {
        ShowErrorMessage("CreateFontData : D3DXCreateFontIndirect Failed\n", "Error");
        ReleaseFontData(dwNum);
        return E_FAIL;
      }
    }
  }

  return hRet;
}

HRESULT ReleaseFontData(DWORD dwNum) {
  if(FontMax < dwNum) return E_FAIL;
  RELEASE(render.lpFontObj[dwNum])
  return S_OK;
}

static HRESULT ResetFontData(DWORD dwNum) {
  if(render.lpFontObj[dwNum]) {
    if(FAILED(render.lpFontObj[dwNum]->OnResetDevice())) {
      return E_FAIL;
    }
  }
  
  if(render.lpFontSprite) {
    if(FAILED(render.lpFontSprite->OnResetDevice())) {
      return E_FAIL;
    }
  }
  
  return S_OK;
}

static HRESULT LostFontData(DWORD dwNum) {
  if(render.lpFontObj[dwNum]) {
    if(FAILED(render.lpFontObj[dwNum]->OnLostDevice())) {
      return E_FAIL;
    }
  }
  
  if(render.lpFontSprite) {
    if(FAILED(render.lpFontSprite->OnLostDevice())) {
      return E_FAIL;
    }
  }
  return S_OK;
}


static HRESULT DrawFontDataSub(DWORD dwNum, FLOAT x, FLOAT y, D3DCOLOR color, LONG shade, LPCSTR szStr) {
  HRESULT hRet = E_FAIL;
  if(FontMax < dwNum) return E_FAIL;
  if(render.lpFontObj[dwNum] && render.lpFontSprite) {
    LPD3DXFONT   lpFont = render.lpFontObj[dwNum];
    LPD3DXSPRITE lpSprite = render.lpFontSprite;
    RECT rc = {(LONG)x + shade, (LONG)y + shade, 0, 0};
    lpSprite->Begin(D3DXSPRITE_ALPHABLEND);
    if(shade) {
      lpFont->DrawText(render.lpFontSprite,  szStr, -1, &rc, DT_NOCLIP, 0xFF000000 & color);
      rc.left -= shade;
      rc.top -= shade;
    }
    lpFont->DrawText(render.lpFontSprite,  szStr, -1, &rc, DT_NOCLIP, color);
    lpSprite->End();
    hRet = S_OK;
  }
  return hRet;
}

HRESULT DrawFontData(DWORD dwNum, FLOAT x, FLOAT y, D3DCOLOR color, LONG shade, const char *fmt, ...) {
  HRESULT hRet = E_FAIL;
  char szBuf[4096];
  if(FontMax < dwNum) return E_FAIL;
  if(render.lpFontObj[dwNum]) {
    va_list ap;
    va_start(ap, fmt);
    vsprintf_s(szBuf, sizeof(szBuf), fmt, ap);
    va_end(ap);
    hRet = DrawFontDataSub(dwNum, x, y, color, shade, szBuf);
  }
  return hRet;
}


//------------------------------------------------------------------------------
// FUNCTION
//------------------------------------------------------------------------------

static LPDIRECT3DTEXTURE9 CreateTextureFromFile(LPCSTR szName) {
  HRESULT             hRet;
  LPDIRECT3DTEXTURE9 lpTex = NULL;
  if(Device) {
    D3DXIMAGE_INFO      iinfo;
    
    if( (hRet = D3DXGetImageInfoFromFile(szName, &iinfo)) == D3D_OK) {
      hRet = D3DXCreateTextureFromFileEx(
        Device,
        szName,
        iinfo.Width, iinfo.Height,
        1, 0,
        D3DFMT_A8R8G8B8,
        D3DPOOL_MANAGED,
        D3DX_FILTER_NONE,
        D3DX_FILTER_NONE,
        0xFF000000,
        NULL, NULL,
        &lpTex);
    }
  }
  return lpTex;
}

HRESULT UnloadTexture(LPCSTR szName) {
  HRESULT hRet = S_OK;
  unsigned long crc32;
  TextureInfo *lpdata = render.texinfo;
  
  //Create crc32
  crc32 = calc_crc32((unsigned char *)szName, strlen(szName));
  for(DWORD count = 0; count < TextureMax; count++) {
    if(lpdata->lpTex) {
      if(lpdata->crc32 == crc32) {
        lpdata->nRef--;
        if(lpdata->nRef <= 0) {
          RELEASE(lpdata->lpTex);
          ::ZeroMemory(lpdata, sizeof(TextureInfo));
        }
      }
    }
  }
  
  return hRet;
}

HRESULT LoadTextureFromFile(LPCSTR szName) {
  HRESULT hRet = E_FAIL;
  unsigned long crc32;
  TextureInfo *lpdata = render.texinfo;
  TextureInfo *find   = NULL;
  TextureInfo *mark   = NULL;
  
  if(!szName) return hRet;
  
  //Create crc32
  crc32 = calc_crc32((unsigned char *)szName, strlen(szName));
  
  //scan tex
  for(DWORD count = 0; count < TextureMax; count++) {
    if(lpdata->lpTex) {
      if(lpdata->crc32 == crc32) {
        lpdata->nRef++;
        find = lpdata;
        hRet = S_OK;
        break;
      }
    } else if(!mark) {
      mark = lpdata;
    }
    lpdata++;
  }
  
  //if not found texture then create new texture.
  if(!find) {
    if(mark) {
      mark->lpTex = CreateTextureFromFile(szName);
      if(mark->lpTex) {
        mark->nRef = 1;
        mark->crc32 = crc32;
        hRet = S_OK;
      }
    }
  }
  
  return hRet;
}

HRESULT SetTexture(LPCSTR szName) {
  HRESULT hRet = S_OK;
  unsigned long crc32 = 0;
  TextureInfo *lpdata = render.texinfo;
  if(Device) {
    if(szName) {
      crc32 = calc_crc32((unsigned char *)szName, strlen(szName));
      for(DWORD count = 0; count < TextureMax; count++) {
        if(lpdata->lpTex) {
          if(lpdata->crc32 == crc32) {
            hRet = Device->SetTexture(0, lpdata->lpTex);
            break;
          }
        }
        lpdata++;
      }
    } else {
      hRet = Device->SetTexture(0, NULL);
    }
  }
  return hRet;
}

//------------------------------------------------------------------------------
// FUNCTION (Draw)
//------------------------------------------------------------------------------
LPDIRECT3DTEXTURE9 GetRenderTexture() {
  return render.lpFlipTexture[render.FlipNumber % FlipTextureMax];
}

LPDIRECT3DSURFACE9 GetRenderSurface() {
  return render.lpFlipSurface[render.FlipNumber % FlipTextureMax];
}

HRESULT SetRenderTarget() {
  HRESULT hRet = E_FAIL;
  if(Device) {
    hRet = Device->SetRenderTarget(0, render.lpFlipSurface[render.FlipNumber % FlipTextureMax]);
    if(hRet != D3D_OK) {
      ShowErrorMessage("CreateResource : SetRenderTarget(render.lpFlipSurface)\n", "Error");
      return hRet;
    }
  }
  return hRet;
}

HRESULT SetRenderTargetBackBuffer() {
  HRESULT hRet = E_FAIL;
  if(Device) {
    //Set Render Target Texture
    hRet = Device->SetRenderTarget(0, render.lpBackSurface);
    if(hRet != D3D_OK) {
      ShowErrorMessage("Flip : SetRenderTarget : Failed\n", "Error");
      return hRet;
    }
    
    //Set Render Target Texture ZBuffer
    hRet = Device->SetDepthStencilSurface(render.lpBackZBuffer);
    if(hRet != D3D_OK) {
      ShowErrorMessage("Flip : SetDepthStencilSurface : Failed\n", "Error");
      return hRet;
    }
  }
  return hRet;
}
  
HRESULT SetDepthStencilSurface() {
  HRESULT hRet = E_FAIL;
  if(Device) {
    hRet = Device->SetDepthStencilSurface(render.lpFlipZBuffer[render.FlipNumber % FlipTextureMax]);
    if(hRet != D3D_OK) {
      ShowErrorMessage("SetDepthStencilSurface : SetDepthStencilSurface(render.lpFlipZBuffer)\n", "Error");
      return hRet;
    }
  }
  return hRet;
}

//#define _DEBUG_RENDER_STATE_ 1

HRESULT SetRenderState(DWORD dwBits) {
  HRESULT hRet = E_FAIL;
  
#ifdef _DEBUG_RENDER_STATE_
  printf("SetRenderState : %08X : ", dwBits);
#endif
  
  //Records
  render.dwRecordBits = dwBits;

  if(Device) {
    //Blend
    if(dwBits & BlendNormal) {
#ifdef _DEBUG_RENDER_STATE_
      printf("BlendNormal:");
#endif
      hRet = Device->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
      hRet = Device->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA );
      hRet = Device->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
    }
    
    if(dwBits & BlendAdd) {
#ifdef _DEBUG_RENDER_STATE_
      printf("BlendAdd:");
#endif
      hRet = Device->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
      hRet = Device->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA );
      //hRet = Device->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_DESTALPHA );
      hRet = Device->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );
    }
    
    //Tss
    if(dwBits & TssNormal) {
#ifdef _DEBUG_RENDER_STATE_
      printf("TssNormal:");
#endif
      hRet = Device->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
      hRet = Device->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
      hRet = Device->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
    }
    
    //Wrap Mode
    if(0) {
      
    } else {
#ifdef _DEBUG_RENDER_STATE_
      printf("Wrap:");
#endif
      hRet = Device->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
      hRet = Device->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
    }
    
    //Filter
    if(dwBits & FilterPoint) {
#ifdef _DEBUG_RENDER_STATE_
      printf("FilterPoint:");
#endif
      hRet = Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_POINT);
      hRet = Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_POINT);
      hRet = Device->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_POINT);
    }
    
    if(dwBits & FilterLinear) {
#ifdef _DEBUG_RENDER_STATE_
      printf("FilterLinear:");
#endif
      hRet = Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
      hRet = Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
      hRet = Device->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
    }
    
    //ZBuffer
    if(dwBits & ZBufferOn) {
      hRet = Device->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
      hRet = Device->SetRenderState(D3DRS_ZENABLE, TRUE);
    } else {
      hRet = Device->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
      hRet = Device->SetRenderState(D3DRS_ZENABLE, FALSE);
    }
    
    //Culling
    if(dwBits & CullOn) {
      hRet = Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
    } else {
      hRet = Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
    }
    
    //Fog
    if(dwBits & FogOn) {
      hRet = Device->SetRenderState(D3DRS_FOGENABLE, TRUE);
    } else {
      hRet = Device->SetRenderState(D3DRS_FOGENABLE, FALSE);
    }
    
    //Light
    if(dwBits & LightOn) {
      hRet = Device->LightEnable( 0, TRUE );
      hRet = Device->SetRenderState( D3DRS_LIGHTING, TRUE );
    } else {
      hRet = Device->LightEnable( 0, FALSE );
      hRet = Device->SetRenderState( D3DRS_LIGHTING, FALSE );
    }
    
  }
#ifdef _DEBUG_RENDER_STATE_
      printf("\n");
#endif
  return hRet;
}

HRESULT SetFogParam(FLOAT fStart, FLOAT fEnd, D3DCOLOR dwColor) {
  HRESULT hRet   = E_FAIL;
  DWORD  dwStart = *(DWORD*)(&fStart);
  DWORD  dwEnd   = *(DWORD*)(&fEnd);
  if(Device) {
    hRet = Device->SetRenderState(D3DRS_FOGVERTEXMODE, D3DFOG_NONE);
    hRet = Device->SetRenderState(D3DRS_FOGTABLEMODE,  D3DFOG_NONE);
    hRet = Device->SetRenderState(D3DRS_FOGTABLEMODE,  D3DFOG_LINEAR);
    hRet = Device->SetRenderState(D3DRS_FOGSTART, dwStart);
    hRet = Device->SetRenderState(D3DRS_FOGEND,   dwEnd);
    hRet = Device->SetRenderState(D3DRS_FOGCOLOR, dwColor);
  }
  return hRet;
}

HRESULT SetRecordRenderState() {
  return SetRenderState(render.dwRecordBits);
}

DWORD GetRecordRenderState() {
  return render.dwRecordBits;
}

HRESULT SetProjection(FLOAT fFov, FLOAT fAspect, FLOAT fNear, FLOAT fFar) {
  HRESULT hRet = E_FAIL;
  if(Device) {
    D3DXMATRIX  mProj;
    D3DXMatrixPerspectiveFovLH(&mProj, fFov, fAspect, fNear, fFar );
    hRet = Device->SetTransform(D3DTS_PROJECTION, &mProj);
  }
  return hRet;
}

HRESULT SetProjection(D3DXMATRIX &mProj) {
  HRESULT hRet = E_FAIL;
  if(Device) {
    hRet = Device->SetTransform(D3DTS_PROJECTION, &mProj);
  }
  return hRet;
}

HRESULT SetView(D3DXVECTOR3 &vPos, D3DXVECTOR3 &vAt, D3DXVECTOR3 &vUp) {
  HRESULT hRet = E_FAIL;
  if(Device) {
    D3DXMATRIX  mView;
    D3DXMatrixLookAtLH(&mView, &vPos,  &vAt,  &vUp);
    hRet = Device->SetTransform(D3DTS_VIEW, &mView);
  }
  return hRet;
}

HRESULT SetView(D3DXMATRIX &mView) {
  HRESULT hRet = E_FAIL;
  if(Device) {
    hRet = Device->SetTransform(D3DTS_VIEW, &mView);
  }
  return hRet;
}

HRESULT SetWorld(D3DXMATRIX &mWorld) {
  HRESULT hRet = E_FAIL;
  if(Device) {
    hRet = Device->SetTransform(D3DTS_WORLD, &mWorld);
  }
  return hRet;
}

HRESULT SetLight(BOOL b, D3DCOLOR ambient, D3DXCOLOR col, D3DXVECTOR3 &v) {
  HRESULT hRet = E_FAIL;
  
  if(Device) {
    D3DXVECTOR3 vecDir;
    D3DLIGHT9   light;
    if(b) {
      ZeroMemory( &light, sizeof(D3DLIGHT9) );
      light.Diffuse.r  = col.r;
      light.Diffuse.g  = col.g;
      light.Diffuse.b  = col.b;
      light.Diffuse.a  = col.a;
      
      vecDir           = v;
      D3DXVec3Normalize(&vecDir, &vecDir);
      light.Direction  = vecDir;
      
      light.Type       = D3DLIGHT_DIRECTIONAL;
      hRet = Device->SetLight(       0, &light );
      hRet = Device->SetRenderState( D3DRS_AMBIENT, ambient );
      hRet = Device->LightEnable(    0, TRUE );
      hRet = Device->SetRenderState( D3DRS_LIGHTING, TRUE );
    } else {
      hRet = Device->LightEnable(    0, FALSE );
      hRet = Device->SetRenderState( D3DRS_LIGHTING, FALSE );
    }
    
  }
  return hRet;
}

HRESULT SetAmbient(D3DCOLOR amb) {
  HRESULT hRet = E_FAIL;
  if(Device) {
    hRet = Device->SetRenderState( D3DRS_AMBIENT, amb );
  }
  return hRet;
}

HRESULT DrawRect2D(FLOAT x1, FLOAT y1, FLOAT x2, FLOAT y2, D3DCOLOR color, FLOAT u1, FLOAT v1, FLOAT u2, FLOAT v2) {
  HRESULT hRet=  E_FAIL;
  if(Device) {
    x1 -= 0.5f;
    y1 -= 0.5f;
    x2 -= 0.5f;
    y2 -= 0.5f;
    
    D4_DIFF_TEX vp[6] = {
        {x1, y1, 0.5f, 1.0f, color, u1, v1},
        {x2, y1, 0.5f, 1.0f, color, u2, v1},
        {x1, y2, 0.5f, 1.0f, color, u1, v2},
        {x2, y1, 0.5f, 1.0f, color, u2, v1},
        {x2, y2, 0.5f, 1.0f, color, u2, v2},
        {x1, y2, 0.5f, 1.0f, color, u1, v2},
    };
    hRet = Device->SetFVF(D4_DIFF_TEX::FVF);
    hRet = Device->DrawPrimitiveUP(D3DPT_TRIANGLELIST, 2, vp, sizeof(D4_DIFF_TEX));
  }
  return hRet;
}

HRESULT DrawRect2D(FLOAT x1, FLOAT y1, FLOAT x2, FLOAT y2, D3DCOLOR color) {
  /*
  HRESULT hRet = E_FAIL;
  if(Device) {
    x1 -= 0.5f;
    y1 -= 0.5f;
    x2 -= 0.5f;
    y2 -= 0.5f;
    
    D4_DIFF_TEX vp[6] = {
        {x1, y1, 0.5f, 1.0f, color, 0.0f, 0.0f},
        {x2, y1, 0.5f, 1.0f, color, 1.0f, 0.0f},
        {x1, y2, 0.5f, 1.0f, color, 0.0f, 1.0f},
        {x2, y1, 0.5f, 1.0f, color, 1.0f, 0.0f},
        {x2, y2, 0.5f, 1.0f, color, 1.0f, 1.0f},
        {x1, y2, 0.5f, 1.0f, color, 0.0f, 1.0f},
    };
    hRet = Device->SetFVF(D4_DIFF_TEX::FVF);
    hRet = Device->DrawPrimitiveUP(D3DPT_TRIANGLELIST, 2, vp, sizeof(D4_DIFF_TEX));
  }
  
  return hRet;
  */
  
  return DrawRect2D(x1,  y1,  x2,  y2,  color, 0.0f, 0.0f, 1.0f, 1.0f);
}

HRESULT DrawRectCenter2D(FLOAT x, FLOAT y, FLOAT w, FLOAT h, D3DCOLOR color) {
  return DrawRect2D(x - w, y - h, x + w, y + h, color, 0.0f, 0.0f, 1.0f, 1.0f);
}

//------------------------------------------------------------------------------
// BATCH DRAW
//------------------------------------------------------------------------------
static std::vector<D4_DIFF_TEX> vBatchDraw2D;


HRESULT DrawBatchRect2DSub(FLOAT x1, FLOAT y1, FLOAT x2, FLOAT y2, D3DCOLOR color, FLOAT u1, FLOAT v1, FLOAT u2, FLOAT v2) {
  HRESULT hRet=  E_FAIL;
  if(Device) {
    x1 -= 0.5f;
    y1 -= 0.5f;
    x2 -= 0.5f;
    y2 -= 0.5f;
    
    D4_DIFF_TEX vp[6] = {
      {x1, y1, 0.5f, 1.0f, color, u1, v1},
      {x2, y1, 0.5f, 1.0f, color, u2, v1},
      {x1, y2, 0.5f, 1.0f, color, u1, v2},
      {x2, y1, 0.5f, 1.0f, color, u2, v1},
      {x2, y2, 0.5f, 1.0f, color, u2, v2},
      {x1, y2, 0.5f, 1.0f, color, u1, v2},
    };
    
    vBatchDraw2D.push_back(vp[0]);
    vBatchDraw2D.push_back(vp[1]);
    vBatchDraw2D.push_back(vp[2]);
    vBatchDraw2D.push_back(vp[3]);
    vBatchDraw2D.push_back(vp[4]);
    vBatchDraw2D.push_back(vp[5]);
  }
  return hRet;
}


HRESULT DrawBatchRect2D(FLOAT x1, FLOAT y1, FLOAT x2, FLOAT y2, D3DCOLOR color) {
  return DrawBatchRect2DSub( x1,  y1,  x2,  y2,  color, 0.0f, 0.0f, 1.0f, 1.0f);
}

HRESULT DrawBatchRectCenter2D(FLOAT x, FLOAT y, FLOAT w, FLOAT h, D3DCOLOR color) {
  return DrawBatchRect2DSub(x - w, y - h, x + w, y + h, color, 0.0f, 0.0f, 1.0f, 1.0f);
}

HRESULT DrawBatchRect2D(FLOAT x1, FLOAT y1, FLOAT x2, FLOAT y2, D3DCOLOR color, FLOAT u1, FLOAT v1, FLOAT u2, FLOAT v2) {
  return DrawBatchRect2DSub( x1,  y1,  x2,  y2, color,  u1,  v1,  u2,  v2);
}

HRESULT DrawBatchRectCenter2D(FLOAT x, FLOAT y, FLOAT w, FLOAT h, D3DCOLOR color, FLOAT u1, FLOAT v1, FLOAT u2, FLOAT v2)  {
  return DrawBatchRect2D(x - w, y - h, x + w, y + h, color,  u1,  v1,  u2,  v2);
}


HRESULT PresentBatchDraw2D() {
  HRESULT hRet = E_FAIL;
  if(Device) {
    if(!vBatchDraw2D.empty()) {
      hRet = Device->SetFVF(D4_DIFF_TEX::FVF);
      hRet = Device->DrawPrimitiveUP(D3DPT_TRIANGLELIST, vBatchDraw2D.size() / 3, &(vBatchDraw2D[0]), sizeof(D4_DIFF_TEX));
    }
  }
  vBatchDraw2D.clear();
  return hRet;
}


//------------------------------------------------------------------------------
// DRAW Line 
//------------------------------------------------------------------------------
static std::vector<D4_DIFF_TEX> vBatchLineDraw2D;

HRESULT DrawBatchLine2D(D3DXVECTOR2 &p1, D3DXVECTOR2 &p2, D3DCOLOR color1, D3DCOLOR color2) {
  HRESULT hRet = E_FAIL;
  if(Device) {
    D4_DIFF_TEX vp[2] = {
        {p1.x,  p1.y, 0.5f, 1.0f, color1, 0.0f, 0.0f},
        {p2.x,  p2.y, 0.5f, 1.0f, color2, 1.0f, 1.0f},
    };
    vBatchLineDraw2D.push_back(vp[0]);
    vBatchLineDraw2D.push_back(vp[1]);
  }

  return hRet;
}

HRESULT PresentBatchLine2D() {
  HRESULT hRet = E_FAIL;
  if(Device) {
    if(!vBatchLineDraw2D.empty()) {
      hRet = Device->SetFVF(D4_DIFF_TEX::FVF);
      hRet = Device->DrawPrimitiveUP(D3DPT_LINELIST, vBatchLineDraw2D.size() / 2, &(vBatchLineDraw2D[0]), sizeof(D4_DIFF_TEX));
    }
  }
  vBatchLineDraw2D.clear();
  return hRet;
}

//------------------------------------------------------------------------------
// DRAW Line 
//------------------------------------------------------------------------------
static std::vector<D3_DIFF_TEX> vBatchLineDraw3D;

HRESULT DrawBatchLine3D(D3DXVECTOR3 &p1, D3DXVECTOR3 &p2, D3DCOLOR color1, D3DCOLOR color2) {
  HRESULT hRet = E_FAIL;
  if(Device) {
    D3_DIFF_TEX vp[2] = {
        {p1.x,  p1.y,  p1.z, color1, 0.0f, 0.0f},
        {p2.x,  p2.y,  p2.z, color2, 1.0f, 1.0f},
    };
    vBatchLineDraw3D.push_back(vp[0]);
    vBatchLineDraw3D.push_back(vp[1]);
  }

  return hRet;
}

HRESULT PresentBatchLine3D() {
  HRESULT hRet = E_FAIL;
  if(Device) {
    if(!vBatchLineDraw3D.empty()) {
      hRet = Device->SetFVF(D3_DIFF_TEX::FVF);
      hRet = Device->DrawPrimitiveUP(D3DPT_LINELIST, vBatchLineDraw3D.size() / 2, &(vBatchLineDraw3D[0]), sizeof(D3_DIFF_TEX));
    }
  }
  vBatchLineDraw3D.clear();
  return hRet;
}


//------------------------------------------------------------------------------
// FUNCTION (MESH)
//------------------------------------------------------------------------------
struct MeshDrawInfo {
  Matrix mLocal;
};

static std::vector<MeshDrawInfo> vmdi;

static HRESULT ReleaseMesh(Mesh *w) {
  if(w) {
    if(w->lpMeshTexture) {
      for(DWORD i = 0 ; i < w->dwMaterialCount ; i++) {
        RELEASE(w->lpMeshTexture[i]);
      }
      SAFE_DELETE_ARRAY(w->lpMeshTexture);
    }
    SAFE_DELETE_ARRAY(w->lpMeshMaterial);
    RELEASE(w->lpMesh);
  }
  
  return S_OK;
}

static HRESULT CreateMeshFromFile(LPCSTR szName, Mesh *mesh) {
  HRESULT hRet = E_FAIL;
  if(szName && Device) {
    LPD3DXBUFFER pMaterial = NULL;
    if( FAILED ( D3DXLoadMeshFromX(szName, D3DXMESH_MANAGED, Device, NULL, &pMaterial, NULL, &mesh->dwMaterialCount, &mesh->lpMesh ) ) ) {
      return hRet;
    }
    
    //Check MATERIAL
    D3DXMATERIAL *pD3DXMaterial = (D3DXMATERIAL *)pMaterial->GetBufferPointer();
    if(mesh->dwMaterialCount <= 0) {
      return hRet;
    }

    mesh->lpMeshMaterial = new D3DMATERIAL9[mesh->dwMaterialCount];
    mesh->lpMeshTexture  = new LPDIRECT3DTEXTURE9[mesh->dwMaterialCount];
    if(mesh->lpMeshTexture) {
      ZeroMemory(mesh->lpMeshTexture, sizeof(LPDIRECT3DTEXTURE9) * mesh->dwMaterialCount);
    }
    if(!mesh->lpMeshMaterial || !mesh->lpMeshTexture) {
      ReleaseMesh(mesh);
      return hRet;
    }
    
    for (DWORD i = 0; i < mesh->dwMaterialCount; i++ ) {
      mesh->lpMeshMaterial[i]         = pD3DXMaterial[i].MatD3D;
      mesh->lpMeshMaterial[i].Ambient = mesh->lpMeshMaterial[i].Diffuse;
      mesh->lpMeshTexture[i]          = NULL;
      if(pD3DXMaterial[i].pTextureFilename != NULL) {
        if(render.bUseMipMap) {
          hRet = D3DXCreateTextureFromFile(Device, pD3DXMaterial[i].pTextureFilename, &(mesh->lpMeshTexture[i]));
        }
        else
        {
          hRet = D3DXCreateTextureFromFileEx( Device,  pD3DXMaterial[i].pTextureFilename, D3DX_DEFAULT, D3DX_DEFAULT, 1, 0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED, D3DX_FILTER_NONE, D3DX_FILTER_NONE, 0xFF000000, NULL, NULL, &(mesh->lpMeshTexture[i]) );
        }
        if(hRet != D3D_OK) {
          mesh->lpMeshTexture[i] = NULL;
        }
      }
    }
    RELEASE(pMaterial);
  }
  return S_OK;
}


HRESULT LoadMeshFromFile(DWORD dwNum, LPCSTR szName) {
  if(dwNum > MeshMax) return E_FAIL;
  Mesh *m = &(render.meshobj[dwNum]);
  ReleaseMesh(m);
  if( FAILED( CreateMeshFromFile(szName, m)  ) ) {
    ReleaseMesh(m);
    return E_FAIL;
  }
  
  return S_OK;
}

HRESULT UnloadMesh(DWORD dwNum) {
  if(dwNum > MeshMax) return E_FAIL;
  return ReleaseMesh(&render.meshobj[dwNum]);
}

HRESULT DrawMesh(D3DXMATRIX &m) {
  HRESULT hRet = E_FAIL;
  MeshDrawInfo temp;
  temp.mLocal = m;
  vmdi.push_back(temp);
  return hRet;
}

HRESULT DrawMesh(D3DXVECTOR3 &pos, D3DXVECTOR3 &rotate, D3DXVECTOR3 &scale) {
  HRESULT hRet = E_FAIL;
  Matrix mRotate;
  Matrix mScale;
  MeshDrawInfo temp;
  
  MatrixTrans(&temp.mLocal, pos.x, pos.y, pos.z);
  MatrixRotateAxis(&mRotate, &Vector3(rotate.x, rotate.y, rotate.z), 1.0f);
  MatrixScalling(&mScale, scale.x, scale.y, scale.z);
  temp.mLocal *= (mScale * mRotate);
  vmdi.push_back(temp);
  
  return hRet;
}



static HRESULT DrawUpMeshSub(DWORD num, HRESULT (*func_draw_mesh)(DWORD, LPD3DXMESH )) {
  HRESULT hRet = E_FAIL;
  Mesh *w = &(render.meshobj[num]);
  if(Device && w->lpMesh) {
    hRet = Device->SetRenderState( D3DRS_ZWRITEENABLE, TRUE);
    hRet = Device->SetRenderState( D3DRS_ZENABLE, D3DZB_TRUE);
    for (DWORD i = 0 ; i < w->dwMaterialCount ; i++) {
      Device->SetMaterial( &(w->lpMeshMaterial[i]) );
      Device->SetTexture( 0, w->lpMeshTexture[i] );
      hRet = func_draw_mesh(i, w->lpMesh);
    }
  }
  return hRet;
}

static HRESULT callback_simple_mesh_draw(DWORD i, LPD3DXMESH lpMesh) {
  return lpMesh->DrawSubset(i);
}

static HRESULT callback_batch_mesh_draw(DWORD i, LPD3DXMESH lpMesh) {
  HRESULT hRet = S_OK;
  std::vector<MeshDrawInfo>::iterator ite = vmdi.begin();
  while(ite != vmdi.end()) {
    SetWorld(ite->mLocal);
    hRet |= lpMesh->DrawSubset(i);
    ite++;
  }
  return hRet;
}

HRESULT DrawUpMesh(DWORD num) {
  return DrawUpMeshSub(num, callback_simple_mesh_draw);
}

HRESULT PresentMesh(DWORD dwNum) {
  HRESULT hRet = E_FAIL;
  if(!vmdi.empty()) {
    hRet = DrawUpMeshSub(dwNum, callback_batch_mesh_draw);
  }
  vmdi.clear();
  return hRet;
}


//------------------------------------------------------------------------------
// FUNCTION (EFFECT)
//------------------------------------------------------------------------------
HRESULT DrawEffectGaussianBlur(D3DCOLOR color) {
  HRESULT hRet = E_FAIL;
  if(Device) {
    LPDIRECT3DTEXTURE9 lpTexture = GetRenderTexture();
    if(lpTexture) {
      
      //Set Texture
      hRet = Device->SetTexture(0, lpTexture);
      if(hRet != D3D_OK) {
        ShowErrorMessage("DrawEffectGaussianBlur : Failed : SetTexture\n", "Error");
        ::PostQuitMessage(0);
        return hRet;
      }

      //Set Render Target (Effect)
      hRet = Device->SetRenderTarget(0, render.lpEffectSurface);
      if(hRet != D3D_OK) {
        ShowErrorMessage("DrawEffectGaussianBlur : SetRenderTarget(render.lpEffectSurface)\n", "Error");
        return hRet;
      }
      
      //ZBuffer is OFF
      hRet = Device->SetDepthStencilSurface(NULL);
      if(hRet != D3D_OK) {
        ShowErrorMessage("DrawEffectGaussianBlur : SetDepthStencilSurface(NULL)\n", "Error");
        return hRet;
      }
      
      //Clear
      hRet = Device->Clear(0, NULL, D3DCLEAR_TARGET, 0, 1.0f, 0);
      if(hRet != D3D_OK) {
        ShowErrorMessage("DrawEffectGaussianBlur : Clear(render.lpEffectSurface)\n", "Error");
        return hRet;
      }
      
      //Get Record Render State
      DWORD tmp = GetRecordRenderState();
      
      //Set Render State
      //printf("//Set Render State START\n");
      SetRenderState(BlendNormal | TssNormal | FilterPoint);
      //printf("//Set Render State END\n");
      
      
      //Calc Texture UV
      DWORD dwZoomX = 2;
      DWORD dwZoomY = 2;
      FLOAT  fZoomU = 1.0f / (FLOAT)dwZoomX;
      FLOAT  fZoomV = 1.0f / (FLOAT)dwZoomY;
      
      //Draw Effect Surface
      DrawRect2D(0, 0, (FLOAT)(render.d3dpp.BackBufferWidth / dwZoomX), (FLOAT)(render.d3dpp.BackBufferHeight / dwZoomY), 0xFFFFFFFF);
      
      //Restore Render Target
      SetRenderTarget();
      SetDepthStencilSurface();
      
      //Set Effect Texture
      hRet = Device->SetTexture(0, render.lpEffectTexture);
      if(hRet != D3D_OK) {
        ShowErrorMessage("DrawEffectGaussianBlur : Failed : SetTexture(render.lpEffectTexture)\n", "Error");
        ::PostQuitMessage(0);
        return hRet;
      }
      
      //Set Render State
      //printf("//Set Render State START\n");
      SetRenderState(BlendAdd | TssNormal | FilterLinear);
      //printf("//Set Render State END\n");
      
      //Render Texture
      hRet = DrawRect2D(0, 0, (FLOAT)render.d3dpp.BackBufferWidth, (FLOAT)render.d3dpp.BackBufferHeight, color, 0.0f, 0.0f, fZoomU, fZoomV);
      
      //Restore Render State
      hRet = SetRenderState(tmp);

    }
  }
  return hRet;
}



//------------------------------------------------------------------------------
// FUNCTION (SYSTEM)
//------------------------------------------------------------------------------

HRESULT CreateResource();

HRESULT TermResource() {
  int i;
  //Release Effect
  RELEASE(render.lpEffectSurface);
  RELEASE(render.lpEffectTexture);
  
  //Release Render Target
  for(i = 0 ; i < FlipTextureMax; i++) {
    RELEASE(render.lpFlipZBuffer[i]);
    RELEASE(render.lpFlipSurface[i]);
    RELEASE(render.lpFlipTexture[i]);
  }
  RELEASE(render.lpBackZBuffer);
  RELEASE(render.lpBackSurface);
  
  return S_OK;
}

HRESULT TermDirectX() {
  DWORD i;
  
  //Release Mesh
  for(i = 0 ; i < MeshMax; i++) {
    UnloadMesh(i);
  }
  
  //Release FontSprite
  RELEASE(render.lpFontSprite)
  
  //Release Font
  for(i = 0 ; i < FontMax; i++) {
    RELEASE(render.lpFontObj[i]);
  }
  
  //Release Load Texture
  for(i = 0 ; i < TextureMax; i++) {
    TextureInfo *lpdata = &(render.texinfo[i]);
    RELEASE(lpdata->lpTex);
  }
  ::ZeroMemory(render.texinfo, sizeof(render.texinfo));
  
  TermResource();
  
  //Release Device
  RELEASE(Device);
  
  //Release D3D Object
  RELEASE(render.lpD3D);
  return S_OK;
}


LPDIRECT3DDEVICE9 GetDevice() {
  return Device;
}

LPCSTR DisplayDriverName() {
  return render.d3daid.Driver;
}

LPCSTR DisplayDescriptionName() {
  return render.d3daid.Description;
}

HRESULT BeginScene() {
  HRESULT hRet = E_FAIL;
  if(Device) {
    hRet = Device->BeginScene();
  }
  return hRet;
}


HRESULT OnLostShader();
HRESULT OnRestoreShader();

static HRESULT OnLost() {
  HRESULT hRet = E_FAIL;
  
  //Lost Shader
  if( FAILED( hRet = OnLostShader()) ) {
    return hRet;
  }
  
  //Lost Font
  for(int i = 0 ; i < FontMax; i++) {
    if(LostFontData(i) != S_OK) {
      return E_FAIL;
    }
  }
  
  if( FAILED(hRet = TermResource()) ) {
    return hRet;
  }
  
  return hRet;
}

static HRESULT OnRestore() {
  HRESULT hRet = E_FAIL;
  hRet = CreateResource();
  
  //Reset Font
  for(int i = 0 ; i < FontMax; i++) {
    ResetFontData(i);
  }
  
  //Restore Shader
  if( FAILED( hRet = OnRestoreShader()) ) {
    return hRet;
  }
  
  return hRet;
}

static HRESULT HandlePresentParam(HRESULT hError, int nRecursive) {
  HRESULT hRet = E_FAIL;
  switch(hError) {
  case D3DERR_DRIVERINTERNALERROR:
    ShowErrorMessage("HandlePresentParam : Failed : D3DERR_DRIVERINTERNALERROR\n", "Error");
    PostQuitMessage(0);
    break;

  case D3DERR_DEVICELOST: {
    if(nRecursive == 0) {
      hRet = HandlePresentParam(Device->TestCooperativeLevel(), 1);
      break;
    } else {
      ::Sleep(LostSleepTime);
      break;
    }
    break;
  }

  case D3DERR_DEVICENOTRESET: {
    DWORD nRepeat = 0;
    while( OnLost() ) {
      nRepeat++;
      if(nRepeat > MaxLostRepeat) {
        ShowErrorMessage("HandlePresentParam : Failed : OnLost\n", "Error");
        ::PostQuitMessage(0);
        hRet = E_FAIL;
        return hRet;
      }
    }
    
    nRepeat = 0;

    hRet = Device->Reset(&render.d3dpp);
    if(hRet != D3D_OK) {
      ShowErrorMessage("HandlePresentParam : Failed : Reset\n", "Error");
      ::PostQuitMessage(0);
      break;
    }

    hRet = OnRestore();
    if(hRet != D3D_OK) {
      ShowErrorMessage("HandlePresentParam : Failed : OnRestore\n", "Error");
      ::PostQuitMessage(0);
      break;
    }
    break;
  }
  default:
    break;
  }
  return hRet;
}

HRESULT EndScene() {
  HRESULT hRet = E_FAIL;
  if(Device) {
    hRet = Device->EndScene();
  }
  return hRet;
}

HRESULT Flip(D3DCOLOR color) {
  HRESULT hRet = E_FAIL;
  if(Device) {
    if(render.bUseBlur) {
      //BeginScene
      if(BeginScene() == S_OK) {
        
        //SetBackBuffer
        SetRenderTargetBackBuffer();

        //After Render Texture
        hRet = Device->SetTexture(0, render.lpFlipTexture[render.FlipNumber % FlipTextureMax]);
        DrawRect2D(0, 0, ((FLOAT)render.d3dpp.BackBufferWidth), ((FLOAT)render.d3dpp.BackBufferHeight), color);

        //EndScene
        EndScene();
      }
    }
    
    hRet = Device->Present(NULL, NULL, NULL, NULL);
    if(hRet != D3D_OK) {
      hRet = HandlePresentParam(hRet, 0);
    }
  }
  return hRet;
}

HRESULT SetBlur(D3DCOLOR dwColor) {
  render.dwColorBlur = dwColor;
  return S_OK;
}

HRESULT EnableBlur() {
  render.bUseBlur = TRUE;
  return S_OK;
}

HRESULT DisableBlur() {
  render.bUseBlur = FALSE;
  SetRenderTargetBackBuffer();
  return S_OK;
}

static HRESULT ClearNoUseBlur(D3DCOLOR color, BOOL bIsAlpha) {
  HRESULT hRet = E_FAIL;
  //Clear BackBuffer
  if(bIsAlpha) {
    if(render.d3dpp.EnableAutoDepthStencil) {
      hRet = Device->Clear(0, NULL, D3DCLEAR_ZBUFFER, color, 1.0f, 0);
    }
    SetRenderState(FilterLinear | BlendNormal | TssNormal);
    SetTexture(NULL);
    
    if(BeginScene() == S_OK) {
      DrawRect2D(0, 0, ((FLOAT)render.d3dpp.BackBufferWidth), ((FLOAT)render.d3dpp.BackBufferHeight), color);
      EndScene();
    }
    
  } else {
    if(render.d3dpp.EnableAutoDepthStencil) {
      hRet = Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, color, 1.0f, 0);
    } else {
      hRet = Device->Clear(0, NULL, D3DCLEAR_TARGET , color, 1.0f, 0);
    }
  }
  return hRet;
}

static HRESULT ClearUseBlur(D3DCOLOR color) {
  HRESULT hRet = E_FAIL;
  
  //incrase Flip Layer
  render.FlipNumber++;
  
  //Set Render Target Texture
  hRet = Device->SetRenderTarget(0, render.lpFlipSurface[render.FlipNumber % FlipTextureMax]);
  if(hRet != D3D_OK) {
    ShowErrorMessage("Clear : Failed : SetRenderTarget\n", "Error");
    return hRet;
  }
  
  //Set Render Target Texture ZBuffer
  hRet = Device->SetDepthStencilSurface(render.lpFlipZBuffer[render.FlipNumber % FlipTextureMax]);
  if(hRet != D3D_OK) {
    ShowErrorMessage("Clear : Failed : SetDepthStencilSurface\n", "Error");
    return hRet;
  }
  
  if(render.d3dpp.EnableAutoDepthStencil) {
    hRet = Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, color, 1.0f, 0);
    if(hRet != D3D_OK) {
      ShowErrorMessage("Clear : Failed : D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER\n", "Error");
      return hRet;
    }
  } else {
    hRet = Device->Clear(0, NULL, D3DCLEAR_TARGET , color, 1.0f, 0);
    if(hRet != D3D_OK) {
      ShowErrorMessage("Clear : Failed : D3DCLEAR_TARGET\n", "Error");
      return hRet;
    }
  }
  
  if(BeginScene() == S_OK) {
    SetRenderState(FilterLinear | BlendNormal | TssNormal);
      
    //Before Render Texture
    hRet = Device->SetTexture(0, render.lpFlipTexture[(render.FlipNumber - 1) % FlipTextureMax]);
    DrawRect2D(0, 0, ((FLOAT)render.d3dpp.BackBufferWidth), ((FLOAT)render.d3dpp.BackBufferHeight), render.dwColorBlur);
    
    //EndScene
    EndScene();
    SetTexture(NULL);
  } else {
    ShowErrorMessage("Clear : Failed : BeginScene\n", "Error");
    return hRet;
  }
  return hRet;
}

HRESULT Clear(D3DCOLOR color, BOOL bIsAlpha) {
  HRESULT hRet = E_FAIL;
  if(Device) {
    if(!render.bUseBlur) {
      hRet = ClearNoUseBlur(color, bIsAlpha);
    } else {
      hRet = ClearUseBlur(color);
    }
  }
  return hRet;
}

HRESULT StartScene(D3DCOLOR color, BOOL bIsAlpha) {
  HRESULT hRet = E_FAIL;
  Clear(color, bIsAlpha);
  hRet = BeginScene();
  return hRet;
}

HRESULT CreateResource() {
  HRESULT hRet = E_FAIL;
  if(Device) {
    hRet = Device->GetRenderTarget(0, &render.lpBackSurface);
    if(hRet != D3D_OK) {
      ShowErrorMessage("CreateResource : GetRenderTarget(BackSurface)\n", "Error");
      return hRet;
    }
        
    hRet = Device->GetDepthStencilSurface(&render.lpBackZBuffer);
    if(hRet != D3D_OK) {
      ShowErrorMessage("CreateResource : GetDepthStencilSurface(render.lpBackZBuffer)\n", "Error");
      return hRet;
    }
    
    //Create Effect Texture
    hRet = Device->CreateTexture(render.d3dpp.BackBufferWidth, render.d3dpp.BackBufferHeight, 1, D3DUSAGE_RENDERTARGET, render.d3dpp.BackBufferFormat, D3DPOOL_DEFAULT, &render.lpEffectTexture, NULL);
    if(hRet != D3D_OK) {
      ShowErrorMessage("CreateResource : CreateTexture(render.lpEffectTexture)\n", "Error");
      return hRet;
    }
    
    //Get Effect Surface
    hRet = render.lpEffectTexture->GetSurfaceLevel(0, &render.lpEffectSurface);
    if(hRet != D3D_OK) {
      ShowErrorMessage("CreateResource : GetSurfaceLevel(render.lpEffectSurface)\n", "Error");
      return hRet;
    }
    
    //Create Render Target
    for(int i = 0 ; i < FlipTextureMax; i++) {
      hRet = Device->CreateTexture(render.d3dpp.BackBufferWidth, render.d3dpp.BackBufferHeight, 1, D3DUSAGE_RENDERTARGET, render.d3dpp.BackBufferFormat, D3DPOOL_DEFAULT, &render.lpFlipTexture[i], NULL);
      if(hRet != D3D_OK) {
        ShowErrorMessage("CreateResource : CreateTexture(render.lpFlipTexture)\n", "Error");
        return hRet;
      }
      
      hRet = render.lpFlipTexture[i]->GetSurfaceLevel(0, &render.lpFlipSurface[i]);
      if(hRet != D3D_OK) {
        ShowErrorMessage("CreateResource : GetSurfaceLevel(render.lpFlipTexture)\n", "Error");
        return hRet;
      }
      
      D3DSURFACE_DESC tempdesc;
      render.lpFlipSurface[i]->GetDesc(&tempdesc);
      
      //Create ZBuffer
      hRet = Device->CreateDepthStencilSurface(render.d3dpp.BackBufferWidth, render.d3dpp.BackBufferHeight, render.d3dpp.AutoDepthStencilFormat, tempdesc.MultiSampleType, tempdesc.MultiSampleQuality, FALSE, &render.lpFlipZBuffer[i], NULL);
      if(hRet != D3D_OK) {
        ShowErrorMessage("CreateResource : CreateDepthStencilSurface(render.lpFlipZBuffer)\n", "Error");
        return hRet;
      }
      
      hRet = Device->SetRenderTarget(0, render.lpFlipSurface[i]);
      if(hRet != D3D_OK) {
        ShowErrorMessage("CreateResource : SetRenderTarget(render.lpFlipSurface)\n", "Error");
        return hRet;
      }
      
      hRet = Device->SetDepthStencilSurface(render.lpFlipZBuffer[i]);
      if(hRet != D3D_OK) {
        ShowErrorMessage("CreateResource : SetDepthStencilSurface(render.lpFlipSurface)\n", "Error");
        return hRet;
      }

      hRet = Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0, 1.0f, 0);
      if(hRet != D3D_OK) {
        ShowErrorMessage("CreateResource : Clear(render.lpFlipSurface)\n", "Error");
        return hRet;
      }
    }
    
    //Restore Back Buffer
    SetRenderTargetBackBuffer();
    if(render.d3dpp.EnableAutoDepthStencil) {
      hRet = Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0, 1.0f, 0);
    } else {
      hRet = Device->Clear(0, NULL, D3DCLEAR_TARGET , 0, 1.0f, 0);
    }
  }
  return hRet;
}

HRESULT CreateDirectXDevice() {
  HRESULT hRet = E_FAIL;

  if(render.lpD3D) {

    //Create DirectX Device
    ZeroMemory(&render.d3dpp, sizeof(render.d3dpp));
    render.d3dpp.Windowed                = GetIsWindow();
    render.d3dpp.BackBufferCount         = 1;
    render.d3dpp.SwapEffect              = D3DSWAPEFFECT_DISCARD;
    render.d3dpp.BackBufferFormat        = render.d3ddm.Format;
    render.d3dpp.MultiSampleType         = D3DMULTISAMPLE_4_SAMPLES;
    
//    render.d3dpp.PresentationInterval    = D3DPRESENT_INTERVAL_IMMEDIATE;
    render.d3dpp.PresentationInterval    = D3DPRESENT_INTERVAL_DEFAULT;
    render.d3dpp.EnableAutoDepthStencil  = TRUE;
    render.d3dpp.AutoDepthStencilFormat  = D3DFMT_D16;
    //render.d3dpp.AutoDepthStencilFormat  = D3DFMT_D24S8;
    render.d3dpp.BackBufferWidth         = GetWidth();
    render.d3dpp.BackBufferHeight        = GetHeight();
    
    //HAL + MultiSample
    hRet = render.lpD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, GetWindowHandle(), D3DCREATE_HARDWARE_VERTEXPROCESSING, &render.d3dpp, &Device);
    if(hRet == D3D_OK) {
      return hRet;
    }
    
    //HAL + NonMultiSample
    render.d3dpp.MultiSampleType = D3DMULTISAMPLE_NONMASKABLE;
    hRet = render.lpD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, GetWindowHandle(), D3DCREATE_HARDWARE_VERTEXPROCESSING, &render.d3dpp, &Device);
    if(hRet == D3D_OK) {
      return hRet;
    }
   
    //HEL
    hRet = render.lpD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, GetWindowHandle(), D3DCREATE_SOFTWARE_VERTEXPROCESSING, &render.d3dpp, &Device);
    if(hRet == D3D_OK) {
      return hRet;
    }
    
    ShowErrorMessage("CreateDirectXDevice : Can't Create DirectX9 Device\n", "Error");
  }
  return E_FAIL;
}

HRESULT InitDirectX() {
  HRESULT hRet = E_FAIL;
  
  //Init var
  ::ZeroMemory(&render, sizeof(render));
  
  //Initialize Direct Graphics
  render.lpD3D = Direct3DCreate9(D3D_SDK_VERSION);
  if(!render.lpD3D) {
    return hRet;
  }
  
  hRet = render.lpD3D->GetAdapterIdentifier(D3DADAPTER_DEFAULT, 0, &render.d3daid);
  if(hRet != D3D_OK) {
    ShowErrorMessage("GetAdapterIdentifier : Failed\n", "Error");
    return hRet;
  }
  
  hRet = render.lpD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &render.d3ddm);
  if(hRet != D3D_OK) {
    ShowErrorMessage("GetAdapterDisplayMode : Failed\n", "Error");
    return hRet;
  }
  
  hRet = CreateDirectXDevice();
  if(hRet != D3D_OK) {
    return hRet;
  }
  
  return CreateResource();
}

